/*
 * $Id: preference.d,v 1.1.1.1 2006/11/19 07:54:55 kenta Exp $
 *
 * Copyright 2006 Kenta Cho. Some rights reserved.
 */
module abagames.ttn.preference;

private import std.stream;
private import abagames.util.preference;

/**
 * Load/Save/Record a high score table.
 */
public class Preference: abagames.util.preference.Preference {
 public:
  static const int RANKING_NUM = 10;
 private:
  static const int VERSION_NUM = 10;
  static const char[] PREF_FILE_NAME = "ttn.prf";
  int[RANKING_NUM] _highScore;

  public void load() {
    auto File fd = null;
    try {
      fd = new File(PREF_FILE_NAME, FileMode.In);
      int ver;
      fd.read(ver);
      if (ver != VERSION_NUM)
        throw new Error("Wrong version num");
      for(int i = 0; i < RANKING_NUM; i++)
        fd.read(_highScore[i]);
    } catch (Object e) {
      init();
    } finally {
      if (fd)
        if (fd.isOpen())
          fd.close();
    }
  }

  private void init() {
    for(int i = 0; i < RANKING_NUM; i++)
      _highScore[i] = (10 - i) * 10000;
  }

  public void save() {
    auto File fd = new File(PREF_FILE_NAME, FileMode.OutNew);
    fd.write(VERSION_NUM);
    for(int i = 0; i < RANKING_NUM; i++)
      fd.write(_highScore[i]);
    fd.close();
  }

  public void recordResult(int score) {
    for(int i = 0; i < RANKING_NUM; i++) {
      if (score > _highScore[i]) {
        for (int j = RANKING_NUM - 1; j >= i + 1; j--) {
          _highScore[j] = _highScore[j - 1];
        }
        _highScore[i] = score;
        return;
      }
    }
  }

  public int[] highScore() {
    return _highScore;
  }
}
